/* noise.c ************************************************\
*                                                          *
*  Higher order Godunov-type scheme for 2D Euler equations *
*             in axisymmetrical formulation                *
*                   on Cartesian grid                      *
*     applied to aeroacoustic problem: the wave issued	   *
*            out of duct into the atmosphere               *
*                                                          *
*  24 November 2000,                                       *
*  author: Andrei A. Chernousov                            *
*  e-mail: <andrei99@iname.com>                            *
*  web: http://www.geocities.com/andrei_chernousov         *
*                                                          *
***********************************************************/

typedef float real;  // double;

/* Features:
  piecewise-parabolic spatial approximation of solution within cell
  linearized characteristic-based approximate Riemann solver (LCS)
  three step Runge-Kutta advancing in time
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include <time.h>
time_t t;

#include <signal.h>

// continuation flag (to be changed by SIGINT handler below)
int continFlag = 1;

// SIGINT handler - handles Ctrl-C key strike
void termHandler( int sigType )
{
   if( continFlag-- ); // 1 -> 0
   puts( "User break, saving arrays and terminating..." );
}

   // combinations of the ratio of specific heats
#define K     1.4
#define K_1   0.4
#define _K_1 (-0.4)
#define _05K  0.7
#define _025K 0.35
#define K_K_1 3.5
#define K_1_2 0.2
#define _1_2_K_1 1.25

  // constants for the Runge-Kutta scheme
#define _0_25  0.25000000000000000
#define _0_33  0.33333333333333333
#define _0_66  0.66666666666666667
#define _0_75  0.75000000000000000

unsigned
   LEN_,  // number of cells in 1D part of the duct
   LENN_, // LEN_+1
   LEN,   // number of 2D cells in x-direction
   HIG,   // number of 2D cells in y-direction
   LENN,  // LEN+1
   HIGG,  // HIG+1
   LEN1,  // number of 2D cells in inside length of the duct
   HIG1,  // number of 2D cells in inner radius of the duct
   LEN2,  // number of 2D cells in outside length of the duct
   HIG2;  // number of 2D cells in outer radius of the duct

real

   // 2D
   **U1,  **U2,  **U3,  **U4,  // arrays for conservative variables
   **U1p, **U2p, **U3p, **U4p, // - final or intermediate states

   **U1_, **U2_, **U3_, **U4_, // pointers for substitution

   **xU1, **xU2, **xU3, **xU4, // values at cells' boundaries
   **U1x, **U2x, **U3x, **U4x,
   **yU1, **yU2, **yU3, **yU4,
   **U1y, **U2y, **U3y, **U4y,

   // 1D
   *UU1,   *UU2,  *UU3,
   *UU1p,  *UU2p, *UU3p,

   *UU1_,  *UU2_, *UU3_,

   *xUU1, *xUU2, *xUU3,
   *UU1x, *UU2x, *UU3x,

   u1, u2, u3, u4,

   // radii
   *r, *rx, *ry,   r_, // Sum( r[j] )

	// primitive variables
   R, U, V, P, C, R_, U_, V_, P_, C_, _R, _U, _V, _P, _C,

   P_1, P_2, R_1, C_1,
   R_2, U_1, U_2, C_2,   PRC,

   P0, P1, R0, I0, Imax, I, C0,  // for initial wave

	// for the characteristic procedure of flux vector evaluation
   C_p, C_m, C_o, C_p_, C_m_, C_o_,
   _jo, _jp, _jm, jo_, jp_, jm_,
   al_p, al_m, al_o,
   J_p, J_m, J_o,

   // x, y spacings and time step
   deltaX, deltaY, deltaT, deltaT_X, deltaT_Y;

unsigned
   step,     // current step
   numstep,  // total number of steps
   outstep,  // number of steps between intermediate outputs
   mem,      // memory in bytes for 2D arrays

   i_start, i_end;

char
   Stage,  // current stage: 1, 2 or 3
   Answer; // do we need to load initial conditions from files (1/0) ?

	// pointers to file handles
FILE *pF,

	 *pFs; // for probes
unsigned i_s, j_s;

  // alias names for fluxes
  // 2Dx
#define Fx1 xU1  // mass flux along X ( RU )
#define Fx2 xU2  // x-momentum flux along X ( RUU+P )
#define Fx3 xU3  // y-momentum flux along X ( RUV )
#define Fx4 xU4  // energy flux along X ( U(RE+P) )
  // 2Dy
#define Fy1 yU1  // mass flux along Y ( RV )
#define Fy2 yU2  // x-momentum flux along Y ( RVU )
#define Fy3 yU3  // y-momentum flux along Y ( RVV+P )
#define Fy4 yU4  // energy flux along Y ( V(RE+P) )
  // 1D
#define FFx1 xUU1
#define FFx2 xUU2
#define FFx3 xUU3


/************
*  ARRAY2D  *   allocates memory for 2D ararys
************/
real **Array2D( unsigned columns, unsigned rows )
{
real **x;
unsigned i;

   if( (x=(real **)malloc(columns*sizeof(real*))) == NULL ) {
	  puts( "Cannot allocate memory" );
	  exit( -1 );
   }
   for( i = 0; i < columns; i++ )
	  if( (x[i]=(real *)malloc(rows*sizeof(real))) == NULL ) {
		puts( "Cannot allocate memory" );
		exit( -1 );
	  }
   return x;

} // end Array2D()


/************
*  Array1D  *   and for 1D ararys
************/
real *Array1D( unsigned size )
{
real *x;

   if( (x=(real *)malloc(size*sizeof(real))) == NULL ) {
	  puts( "Cannot allocate memory" );
	  exit( -1 );
   }
   return x;

} // end Array1D( )


/***************
*  INITIALIZE  *   perform initializations
***************/
void Initialize( void )
{
unsigned i, j;
char str[80];

   // read inits from the file
   if( ( pF = fopen( "noise.ini", "r" ) ) == NULL ) {
	  puts( "Cannot open file \"noise.ini\"" );
	  exit( -1 );
   }
   i = 0;
   do {
	  if ( fgets( str, 79, pF ) != NULL ) {
		 switch( i ) {
			case 0:	P0     = atof( str ); break;
			case 1:	P1     = atof( str ); break;
			case 2:	R0     = atof( str ); break;
			case 3:  i_start= atoi( str ); break;
			case 4:  i_end  = atoi( str ); break;
		    case 5:	LEN_   = atoi( str ); break;
			case 6:	LEN    = atoi( str ); break;
			case 7:	HIG    = atoi( str ); break;
			case 8:	LEN1   = atoi( str ); break;
			case 9:	HIG1   = atoi( str ); break;
			case 10:	LEN2   = atoi( str ); break;
			case 11:  HIG2   = atoi( str ); break;
			case 12: deltaX = atof( str ); break;
			case 13: deltaY = atof( str ); break;
			case 14: deltaT = atof( str ); break;
			case 15: Answer = (char)atoi( str ); break;
			case 16: numstep= atoi( str ); break;
			case 17: outstep= atoi( str ); break;
			case 18: i_s    = atoi( str ); break;
			case 19: j_s    = atoi( str );
			default: break;
		 } // end switch
		 i++;
	  } // end if
	  else {
		puts( "Error reading file \"noise.ini\"" );
		exit( -1 );
	 }
   } while( i < 20 );
   fclose( pF );

   // some other initializations
   deltaT_X = deltaT / deltaX;
   deltaT_Y = deltaT / deltaY;
   LENN = LEN+1;
   HIGG = HIG+1;
   LENN_ = LEN_+1;

   // amount of dynamic memory needed
	 // 2D:
   mem = sizeof( real ) * 8 * ( ( ( LEN+2 ) * ( HIG+2 ) )
							   + ( ( LEN+1 ) *   HIG     )
							   + (   LEN     * ( HIG+1 ) ) );
   printf( "%d bytes of dynamic memory is needed. Allocating...", mem );
   // conservative variables in cells
   U1 = Array2D( LEN+2, HIG+2 );
   U2 = Array2D( LEN+2, HIG+2 );
   U3 = Array2D( LEN+2, HIG+2 );
   U4 = Array2D( LEN+2, HIG+2 );
   U1p= Array2D( LEN+2, HIG+2 );
   U2p= Array2D( LEN+2, HIG+2 );
   U3p= Array2D( LEN+2, HIG+2 );
   U4p= Array2D( LEN+2, HIG+2 );
   // x-boundaries
   xU1 = Array2D( LEN+1, HIG );
   xU2 = Array2D( LEN+1, HIG );
   xU3 = Array2D( LEN+1, HIG );
   xU4 = Array2D( LEN+1, HIG );
   U1x = Array2D( LEN+1, HIG );
   U2x = Array2D( LEN+1, HIG );
   U3x = Array2D( LEN+1, HIG );
   U4x = Array2D( LEN+1, HIG );
   // y-boundaries
   yU1 = Array2D( LEN, HIG+1 );
   yU2 = Array2D( LEN, HIG+1 );
   yU3 = Array2D( LEN, HIG+1 );
   yU4 = Array2D( LEN, HIG+1 );
   U1y = Array2D( LEN, HIG+1 );
   U2y = Array2D( LEN, HIG+1 );
   U3y = Array2D( LEN, HIG+1 );
   U4y = Array2D( LEN, HIG+1 );
	 // 1D:
   // conservative variables in cells
   UU1 = Array1D( LEN_+2 );
   UU2 = Array1D( LEN_+2 );
   UU3 = Array1D( LEN_+2 );
   UU1p= Array1D( LEN_+2 );
   UU2p= Array1D( LEN_+2 );
   UU3p= Array1D( LEN_+2 );
   // x-boundaries
   xUU1 = Array1D( LEN_+1 );
   xUU2 = Array1D( LEN_+1 );
   xUU3 = Array1D( LEN_+1 );
   UU1x = Array1D( LEN_+1 );
   UU2x = Array1D( LEN_+1 );
   UU3x = Array1D( LEN_+1 );
	// radii
   r  = Array1D( HIG+1 ); for( j = 1; j <= HIG; r [j] = (j-0.5) * deltaX, j++ );
   rx = Array1D( HIG   ); for( j = 0; j <  HIG; rx[j] = (j+0.5) * deltaX, j++ );
   ry = Array1D( HIG+1 ); for( j = 0; j <= HIG; ry[j] =  j      * deltaX, j++ );
   //
   for( j = 1, r_ = 0.; j <= HIG1; r_ += r [j], j++ );
   //
   printf( " done!\n" );

   // initial conditions
   if( Answer ) {
	  if ( ( pF = fopen( "backup", "rb") ) == NULL ) {
		puts( "Cannot open file \"backup\"" );
		exit(1);
	  }
	  // 2D
	  for ( i = 1; i <= LEN; i++ ) {
		 fread( &U1[i][1], HIG * sizeof( real ), 1, pF );
		 fread( &U2[i][1], HIG * sizeof( real ), 1, pF );
		 fread( &U3[i][1], HIG * sizeof( real ), 1, pF );
		 fread( &U4[i][1], HIG * sizeof( real ), 1, pF );
	  }
	  // 1D
	  fread( &UU1[1], LEN_ * sizeof( real ), 1, pF );
	  fread( &UU2[1], LEN_ * sizeof( real ), 1, pF );
	  fread( &UU3[1], LEN_ * sizeof( real ), 1, pF );
	  // step = ...
	  fread( &step, sizeof( unsigned ), 1, pF );
	  //
	  fclose( pF );
   }
   else {
	  // 2D
	  for ( i = 1; i <= LEN; i++ ) {
		 for ( j = 1; j <= HIG; j++ ) {
			U1[i][j] = R0;
			U2[i][j] = 0;
			U3[i][j] = 0;
			U4[i][j] = P0 / K_1;
		 }
	  }
	  // 1D
        //
     C0 = sqrt( K * P0 / R0 );
     C  = C0 * pow( P1 / P0, K_1/(K+K) );
     I0 = ( C0 + C0 ) / K_1;
     Imax = 4 * C / K_1 - I0;
        // sinusoidal Riemann invariant initial wave
	  for ( i = 1; i <= LEN_; i++ ) {
       if( i > i_start && i <= i_end ) {
          I = I0 + ( Imax - I0 ) * sin( M_PI * ( i - i_start - 0.5 )
                                             / ( i_end - i_start ) );
          C = 0.25 * K_1 * ( I + I0 );
          U = 0.5 * ( I - I0 );
          P = P0 * pow( C / C0, (K+K)/K_1 );
          R = K * P / ( C * C );
       }
        // far-field
       else {
          P = P0;
          R = R0;
          U = 0.;
       }
		 UU1[i] = R;
		 UU2[i] = R * U;
		 UU3[i] = P / K_1 + 0.5 * R * U * U;
	  }
	  step = 0;
   } // end else()

   // opens probes file in ascii mode
   if( ( pFs = fopen( "probes.dat", "a" ) ) == NULL ) {
	  puts( "Cannot open file \"probes.dat\"" );
	  exit( -1 );
   }

   /* next stage */		 
   Stage = 1;
   U1_ = U1,  U2_ = U2,  U3_ = U3,  U4_ = U4;
   UU1_= UU1, UU2_= UU2, UU3_= UU3;

} // end Initialize()

/********************
* BOUNDARYCONDITONS *   boundary conditions
********************/
void BoundaryConditions( void )
{
register unsigned i, j;

   //-- 2D
   // X
   // right boundary (1)
   for( j = 1; j < HIGG; j++ ) {
	  // reflection on the right
	  U1_[LENN][j] =   U1_[LEN][j];
	  U2_[LENN][j] = - U2_[LEN][j];
	  U3_[LENN][j] =   U3_[LEN][j];
	  U4_[LENN][j] =   U4_[LEN][j];
   }
   // left boundary (2)
   for( j = HIG2 + 1; j < HIGG; j++ ) {
	  // reflection on the left
	  U1_[0][j] =   U1_[1][j];
	  U2_[0][j] = - U2_[1][j];
	  U3_[0][j] =   U3_[1][j];
	  U4_[0][j] =   U4_[1][j];
   }
   // edge of the duct (3)
   for( j = HIG1 + 1; j <= HIG2; j++ ) {
	  // reflection on the left
	  U1_[LEN2][j] =   U1_[LEN2+1][j];
	  U2_[LEN2][j] = - U2_[LEN2+1][j];
	  U3_[LEN2][j] =   U3_[LEN2+1][j];
	  U4_[LEN2][j] =   U4_[LEN2+1][j];
   }
   // inside the duct (4)
   for( j = 1; j <= HIG1; j++ ) {
	  // connection to the 1D duct
	  U1_[LEN1][j] = UU1_[LEN_];
	  U2_[LEN1][j] = UU2_[LEN_];
	  U3_[LEN1][j] = 0.;
	  U4_[LEN1][j] = UU3_[LEN_];
   }

   // Y
   // upper boundary (5)
   for( i = 1; i < LENN; i++ ) {
	  U1_[i][HIGG] =   U1_[i][HIG];
	  U2_[i][HIGG] =   U2_[i][HIG];
	  U3_[i][HIGG] = - U3_[i][HIG];
	  U4_[i][HIGG] =   U4_[i][HIG];
   } // end for()
   // boundary at the outer radius of the duct (6)
   for( i = 1; i <= LEN2; i++ ) {
	  U1_[i][HIG2] =   U1_[i][HIG2+1];
	  U2_[i][HIG2] =   U2_[i][HIG2+1];
	  U3_[i][HIG2] = - U3_[i][HIG2+1];
	  U4_[i][HIG2] =   U4_[i][HIG2+1];
   } // end for()
   // boundary at the inner radius of the duct (7)
   for( i = LEN1+1; i <= LEN2; i++ ) {
	  U1_[i][HIG1+1] =   U1_[i][HIG1];
	  U2_[i][HIG1+1] =   U2_[i][HIG1];
	  U3_[i][HIG1+1] = - U3_[i][HIG1];
	  U4_[i][HIG1+1] =   U4_[i][HIG1];
   } // end for()
   // lower boundary (8)
   for( i = LEN1+1; i < LENN; i++ ) {
	  U1_[i][0] =   U1_[i][1];
	  U2_[i][0] =   U2_[i][1];
	  U3_[i][0] = - U3_[i][1];
	  U4_[i][0] =   U4_[i][1];
   } // end for()

   //-- 1D
	 // left end of 1D duct
   UU1_[0] =   UU1_[1]; // reflective BC
   UU2_[0] = - UU2_[1];
   UU3_[0] =   UU3_[1];
   // right end of 1D duct - not needed
	 // mean values in 1D ghost cell by averaging from the inside of the duct
   for( j = 1, u1 = u2 = u3 = 0; j <= HIG1; j++ ) {
	  u1 += U1_[LEN1+1][j] * r[j];
	  u2 += U2_[LEN1+1][j] * r[j];
	  u3 += U4_[LEN1+1][j] * r[j];
   }
   UU1_[LENN_] = u1 / r_;
   UU2_[LENN_] = u2 / r_;
   UU3_[LENN_] = u3 / r_;

} // end BounaryConditions( )

//--- For monotone piecewise parabolic reconstruction ---
	// constant coefficients
#define BB 4.000000000000000
#define C1 0.333333333333333
#define C2 0.166666666666667

// limiter of [Chakravarthy, Osher]
/*inline*/ real minmod( real x, real y ) {
	if( x*y<=0 ) return 0;
	y = ( (y*=BB) > 0 ) ? y: -y;
	if( x > 0 ) return (x<y) ? x : y;
	return ( (-x<y) ? x: -y );
} // end minmod()

real
   // finite differences of conservative variables
   m1, m2, m3, m4, w1, w2, w3, w4,
   // finite differences of characteristic variables
   M1, M2, M3, M4, W1, W2, W3, W4,
   // modified finite differences of characteristic variables
   _M1, _M2, _M3, _M4, _W1, _W2, _W3, _W4,
   // transformation matrix [S-1], etc...
   S11, S12, S13, S14,
   S21, S22, S23, S24,
   S31, S32, S33, S34,
   S41, S42, S43, S44,
   // inverted transformation matrix [S-1]
   _S11, _S12, _S13, _S14,
   _S21, _S22, _S23, _S24,
   _S31, _S32, _S33, _S34,
   _S41, _S42, _S43, _S44,
   //
   aaa, bbb, ccc, ddd, AAA, BBB, CCC, DDD, EEE;


/***************
*  MATRICES2D  *    fills transformation matrices for 2D
***************/
void Matrices2D( short r )
{
   if( r == 1 ) {
	  aaa = 0.5 * K_1 * ( U * U + V * V );
	  ddd = C * C;
	  AAA = - 1. / ddd;
	  BBB = 1. / ( 2. * C );
	  CCC = 0.5 * AAA * ( U * U + V * V );
   }
   // [S]
   bbb = - K_1 * U;
   S11 = -ddd + aaa;       S12 = bbb;     S13 = -K_1*V; //S14 = K_1;     // S11 = -C*C + K_1*(u*u+v*v)/2; S12 =    - K_1*u; S13 = -K_1*v;	S14 = K_1;
   S21 = -V;             //S22 = 0.;      S23 = 1.;       S24 = 0.;      // S21 = -v;                     S22 =          0; S23 = 1;	    S24 = 0;
   S31 = -(ccc=C*U)+ aaa;  S32 = bbb+C; //S33 = S13;    //S34 = K_1;     // S31 = -c*u + K_1*(u*u+v*v)/2; S32 =  c - K_1*u; S33 = -K_1*v;	S34 = K_1;
   S41 =   ccc     + aaa;  S42 = bbb-C; //S43 = S13;    //S44 = K_1;     // S41 =  c*u + K_1*(u*u+v*v)/2; S42 = -c - K_1*u; S43 = -K_1*v;	S44 = K_1;
   // [S^-1]
   DDD = - 0.5 * AAA * U;
   EEE = - 0.5 * AAA * V;
   _S11 = AAA;                _S13 = - 0.5 * AAA;                 //_S14 = _S13; 							//	    _S11 = - 1 /(c*c);          _S12 = 0;  _S13 = 1/(2*c*c);                               _S14 = 1/(2*c*c);
   _S21 = AAA*U;              _S23 = DDD + BBB;                     _S24 = DDD - BBB;                       //		_S21 = - u /(c*c);          _S22 = 0;  _S23 = u/(2*c*c) + 1/(2*c);                     _S24 = u/(2*c*c) - 1/(2*c);
   _S31 = AAA*V;              _S33 = EEE;                         //_S34 = EEE;                             //		_S31 = - v /(c*c);          _S32 = 1;  _S33 = v/(2*c*c);                               _S34 = v/(2*c*c);
   _S41 = CCC;    _S42 = V;   _S43 = -0.5*_S41 + _1_2_K_1 + U*BBB;  _S44 = _S43 - 2. * U * BBB;                 //		_S41 = -(u*u+v*v)/(2.*c*c); _S42 = v;  _S43 = (u*u+v*v)/(4*c*c) + 1/(2*K_1) + u/(2*c); _S44 = (u*u+v*v)/(4*c*c) + 1/(2*K_1) - u/(2*c); */
} // end Matrices()


/*******************
*  RECONSTRUCTION  *   piecewise-parabolic reconstruction of parameters in cells
*******************/
void Reconstruction( void )
{
register unsigned _i, _j, i, j, i_, j_;
register real k1, k2, k3, k4;

   //-- 2D
   // piecewise-parabolic reconstruction
   for( _i = 0, i = 1, i_ = 2; i < LENN; _i++, i++, i_++ )  {
	  for( _j = 0, j = 1, j_ = 2; j < HIGG; _j++, j++, j_++ ) {
		 // selection
		 if( j <= HIG1 ) { if( i <= LEN1 ) continue; }
		 else if( j <= HIG2 ) { if( i <= LEN2) continue; }
		 //--- X
	    // at cell centroid
		 u1 = U1_[i][j];
		 u2 = U2_[i][j];
		 u3 = U3_[i][j];
		 u4 = U4_[i][j];
		 R =   u1;
		 U =   u2 / R;
		 V =   u3 / R;
		 P = ( u4 - 0.5 * R * ( U * U + V * V ) ) * K_1;
		 C = sqrt( K * P / R );
		 // evaluation of matrices
		 Matrices2D( 1 );
		 // differences
			// forward
		 m1 = U1_[i_][j] - u1;
		 m2 = U2_[i_][j] - u2;
		 m3 = U3_[i_][j] - u3;
		 m4 = U4_[i_][j] - u4;
			// backward
		 w1 = u1 - U1_[_i][j];
		 w2 = u2 - U2_[_i][j];
		 w3 = u3 - U3_[_i][j];
		 w4 = u4 - U4_[_i][j];
		 // differences of characteristic fields
		 M1 = S11*m1 + S12*m2 + S13*m3 + K_1*m4;
		 M2 = S21*m1              + m3         ;
		 M3 = S31*m1 + S32*m2 + S13*m3 + K_1*m4;
		 M4 = S41*m1 + S42*m2 + S13*m3 + K_1*m4;
		 W1 = S11*w1 + S12*w2 + S13*w3 + K_1*w4;
		 W2 = S21*w1          +     w3         ;
		 W3 = S31*w1 + S32*w2 + S13*w3 + K_1*w4; // S33 = S13
		 W4 = S41*w1 + S42*w2 + S13*w3 + K_1*w4; // S43 = S13
		 // modified characteristic differences _W -
		 _M1 = minmod( M1, W1 );
		 _M2 = minmod( M2, W2 );
		 _M3 = minmod( M3, W3 );
		 _M4 = minmod( M4, W4 );
		 _W1 = minmod( W1, M1 );
		 _W2 = minmod( W2, M2 );
		 _W3 = minmod( W3, M3 );
		 _W4 = minmod( W4, M4 );
		 // rightmost interface
		 k1 = _M1 * C1 + _W1 * C2;
		 k2 = _M2 * C1 + _W2 * C2;
		 k3 = _M3 * C1 + _W3 * C2;
		 k4 = _M4 * C1 + _W4 * C2;
		 xU1[ i][_j] = u1 + _S11 * k1		         + _S13 * ( k3 +        k4 );
		 xU2[ i][_j] = u2 + _S21 * k1 		      + _S23 *   k3 + _S24 * k4;
		 xU3[ i][_j] = u3 + _S31 * k1 +        k2 + _S33 * ( k3 +        k4 );
		 xU4[ i][_j] = u4 + _S41 * k1 + _S42 * k2 + _S43 *   k3 + _S44 * k4;
		 // leftmost interface
		 k1 = _M1 * C2 + _W1 * C1;
		 k2 = _M2 * C2 + _W2 * C1;
		 k3 = _M3 * C2 + _W3 * C1;
		 k4 = _M4 * C2 + _W4 * C1;
		 U1x[_i][_j] = u1 - _S11 * k1		         - _S13 * ( k3 +        k4 );
		 U2x[_i][_j] = u2 - _S21 * k1 		      - _S23 *   k3 - _S24 * k4;
		 U3x[_i][_j] = u3 - _S31 * k1 -        k2 - _S33 * ( k3 +        k4 );
		 U4x[_i][_j] = u4 - _S41 * k1 - _S42 * k2 - _S43 *   k3 - _S44 * k4;
		 //--- Y
		 // at cell centroid (V->U)
		 k1 = U; U = V; V = k1;
		 k1 = u2; u2 = u3; u3 = k1;
		 // evaluation of matrices
		 Matrices2D( 2 );
		 // differences
			// upper
		 m1 = U1_[i][j_] - u1;
		 m2 = U3_[i][j_] - u2; // substitution U->V, V->U
		 m3 = U2_[i][j_] - u3;
		 m4 = U4_[i][j_] - u4;
			// lower
		 w1 = u1 - U1_[i][_j];
		 w2 = u2 - U3_[i][_j];
		 w3 = u3 - U2_[i][_j];
		 w4 = u4 - U4_[i][_j];
		 // differences of characteristic fields
		 M1 = S11*m1 + S12*m2 + S13*m3 + K_1*m4;
		 M2 = S21*m1              + m3         ;
		 M3 = S31*m1 + S32*m2 + S13*m3 + K_1*m4;
		 M4 = S41*m1 + S42*m2 + S13*m3 + K_1*m4;
		 W1 = S11*w1 + S12*w2 + S13*w3 + K_1*w4;
		 W2 = S21*w1          +     w3         ;
		 W3 = S31*w1 + S32*w2 + S13*w3 + K_1*w4;
		 W4 = S41*w1 + S42*w2 + S13*w3 + K_1*w4;
		 // modified characteristic differences _W -
		 //   minmod( M, W ) limiter expanded explicitly inline
		 _M1 = minmod( M1, W1 );
		 _M2 = minmod( M2, W2 );
		 _M3 = minmod( M3, W3 );
		 _M4 = minmod( M4, W4 );
		 _W1 = minmod( W1, M1 );
		 _W2 = minmod( W2, M2 );
		 _W3 = minmod( W3, M3 );
		 _W4 = minmod( W4, M4 );
		 // upper interface
		 k1 = _M1 * C1 + _W1 * C2;
		 k2 = _M2 * C1 + _W2 * C2;
		 k3 = _M3 * C1 + _W3 * C2;
		 k4 = _M4 * C1 + _W4 * C2;
		 yU1[_i][ j] = u1 + _S11 * k1		         + _S13 * ( k3 +        k4 );
		 yU3[_i][ j] = u2 + _S21 * k1 		      + _S23 *   k3 + _S24 * k4;
		 yU2[_i][ j] = u3 + _S31 * k1 +        k2 + _S33 * ( k3 +        k4 );
		 yU4[_i][ j] = u4 + _S41 * k1 + _S42 * k2 + _S43 * k3 +   _S44 * k4;
		 // lower interface
		 k1 = _M1 * C2 + _W1 * C1;
		 k2 = _M2 * C2 + _W2 * C1;
		 k3 = _M3 * C2 + _W3 * C1;
		 k4 = _M4 * C2 + _W4 * C1;
		 U1y[_i][_j] = u1 - _S11 * k1	            - _S13 * ( k3 +        k4 );
		 U3y[_i][_j] = u2 - _S21 * k1 		      - _S23 *   k3 - _S24 * k4;
		 U2y[_i][_j] = u3 - _S31 * k1 -        k2 - _S33 * ( k3 +        k4 );
		 U4y[_i][_j] = u4 - _S41 * k1 - _S42 * k2 - _S43 *   k3 - _S44 * k4;
	  } // end for
   } // end for

   //-- 1D
   for( _i = 0, i = 1, i_ = 2; i < LENN_; _i++, i++, i_++ ) {
	  // at cell centroid
	  R =   UU1_[i];
	  U =   UU2_[i] / R;
	  P = ( UU3_[i] - 0.5 * R * U * U ) * K_1;
	  C = sqrt( K * P / R );
	  // inverted transformation matrix [S-1]
	  AAA = 1/(C+C);
	  BBB = AAA / C;
	  CCC = BBB + BBB;
	  DDD = 0.5 * U * U * BBB;
	  _S11 = -(BBB+BBB); _S12 = _S13 = BBB;
	  _S21 = -U*CCC;     _S22 = _S23 = U * BBB;        _S22 += AAA;               _S23 -= AAA;
	  _S31 = -(DDD+DDD); _S32 = _S33 = DDD + _1_2_K_1; _S32 += ( AAA = U * AAA ); _S33 -= AAA;
	  // transformation matrix [S]
	  AAA = K_1_2*U*U;
	  BBB = C*U;
	  CCC = -K_1*U;
	  S11 = -C*C + AAA;  S12 = CCC;
	  S21 = -BBB + AAA;  S22 = CCC+C;
	  S31 =  BBB + AAA;  S32 = CCC-C;
	  // differencing of the conservative variables
		 // forward
	  m1 = UU1_[i_] - UU1_[i];
	  m2 = UU2_[i_] - UU2_[i];
	  m3 = UU3_[i_] - UU3_[i];
		  // backward
	  w1 = UU1_[i] - UU1_[_i];
	  w2 = UU2_[i] - UU2_[_i];
	  w3 = UU3_[i] - UU3_[_i];
	  // finite differences of characteristic variables dW = S * dU
		 // forward
	  W1 = S11*w1 + S12*w2 + K_1*w3;
	  W2 = S21*w1 + S22*w2 + K_1*w3;
	  W3 = S31*w1 + S32*w2 + K_1*w3;
		 // backward
	  M1 = S11*m1 + S12*m2 + K_1*m3;
	  M2 = S21*m1 + S22*m2 + K_1*m3;
	  M3 = S31*m1 + S32*m2 + K_1*m3;
	  // modified characteristic differences _W
	  _M1 = minmod( M1, W1 );
	  _M2 = minmod( M2, W2 );
	  _M3 = minmod( M3, W3 );
	  _W1 = minmod( W1, M1 );
	  _W2 = minmod( W2, M2 );
	  _W3 = minmod( W3, M3 );
	  // parameter vector at
		 // rightmost boundary
	  k1 = _M1 * C1 + _W1 * C2;
	  k2 = _M2 * C1 + _W2 * C2;
	  k3 = _M3 * C1 + _W3 * C2;
	  xUU1[ i] = UU1_[i] + _S11 * k1 + _S12 * k2 + _S13 * k3;
	  xUU2[ i] = UU2_[i] + _S21 * k1 + _S22 * k2 + _S23 * k3;
	  xUU3[ i] = UU3_[i] + _S31 * k1 + _S32 * k2 + _S33 * k3;
		 // leftmost boundary
	  k1 = _M1 * C2 + _W1 * C1;
	  k2 = _M2 * C2 + _W2 * C1;
	  k3 = _M3 * C2 + _W3 * C1;
	  UU1x[_i] = UU1_[i] - _S11 * k1 - _S12 * k2 - _S13 * k3;
	  UU2x[_i] = UU2_[i] - _S21 * k1 - _S22 * k2 - _S23 * k3;
	  UU3x[_i] = UU3_[i] - _S31 * k1 - _S32 * k2 - _S33 * k3;
   } // end for()

} // end Reconstruction()


/*******************
*  BConInterfaces  *   completion of the solution at the outer sides of cell boundaries
*******************/
void BConInterfaces( void )
{
register unsigned i, j;

   //-- 2D
   // X
   // right boundary (1)
   for( j = 0; j < HIG; j++ ) {
	  // reflection from the right
	  U1x[LEN][j] =   xU1[LEN][j];
	  U2x[LEN][j] = - xU2[LEN][j];
	  U3x[LEN][j] =   xU3[LEN][j];
	  U4x[LEN][j] =   xU4[LEN][j];
   } // end for()
   // left boundary (2)
   for( j = HIG2; j < HIG; j++ ) {
	  // reflection from the left
	  xU1[0][j] =   U1x[0][j];
	  xU2[0][j] = - U2x[0][j];
	  xU3[0][j] =   U3x[0][j];
	  xU4[0][j] =   U4x[0][j];
   } // end for()
   // edge of the duct (3)
   for( j = HIG1; j < HIG2; j++ ) {
	  // reflection from the left
	  xU1[LEN2][j] =   U1x[LEN2][j];
	  xU2[LEN2][j] = - U2x[LEN2][j];
	  xU3[LEN2][j] =   U3x[LEN2][j];
	  xU4[LEN2][j] =   U4x[LEN2][j];
   } // end for()
   // inside the duct (4)
   for( j = 0; j < HIG1; j++ ) {
	  // connecction to the 1D duct
	  xU1[LEN1][j] = xUU1[LEN_];
	  xU2[LEN1][j] = xUU2[LEN_];
	  xU3[LEN1][j] = 0.;
	  xU4[LEN1][j] = xUU3[LEN_];
   } // end for()
   // Y
   // upper boundary (5)
   for( i = 0; i < LEN; i++ ) {
	  U1y[i][HIG] =   yU1[i][HIG];
	  U2y[i][HIG] =   yU2[i][HIG];
	  U3y[i][HIG] = - yU3[i][HIG];
	  U4y[i][HIG] =   yU4[i][HIG];
   } // end for()
   // boundary at the outer radius of the duct (6)
   for( i = 0; i < LEN2; i++ ) {
	  yU1[i][HIG2] =   U1y[i][HIG2];
	  yU2[i][HIG2] =   U2y[i][HIG2];
	  yU3[i][HIG2] = - U3y[i][HIG2];
	  yU4[i][HIG2] =   U4y[i][HIG2];
   } // end for()
   // boundary at the inner radius of the duct (7)
   for( i = LEN1; i < LEN2; i++ ) {
	  U1y[i][HIG1] =   yU1[i][HIG1];
	  U2y[i][HIG1] =   yU2[i][HIG1];
	  U3y[i][HIG1] = - yU3[i][HIG1];
	  U4y[i][HIG1] =   yU4[i][HIG1];
   } // end for()
   // lower boundary (8)
   for( i = LEN1; i < LEN; i++ ) {
	  yU1[i][0] =   U1y[i][0];
	  yU2[i][0] =   U2y[i][0];
	  yU3[i][0] = - U3y[i][0];
	  yU4[i][0] =   U4y[i][0];
   } // end for()

   //-- 1D
   // on the left end - reflection
   xUU1[0] =   UU1x[0];
   xUU2[0] = - UU2x[0];
   xUU3[0] =   UU3x[0];
   // on the right end - not needed: fluxes at LENN_ are not evaluated in 1D

} // end BConInterfaces( )


/***********
*  FLUXES  *   fluxes at the boundaries
***********/
void Fluxes( void )
{
register unsigned i, j;
register real RU; // convective normal mass flux

   //-- 2D
   // X-direction
   for( i = 0; i < LENN; i++ ) {
	  for( j = 0; j < HIG; j++ ) {
		 // selection
		 if( j < HIG1 ) { if( i < LEN1 ) continue; }
		 else if( j < HIG2 ) { if( i < LEN2 ) continue; }
		 // left parameters
		 _R =   xU1[i][j];
		 _U =   xU2[i][j] / _R;
		 _V =   xU3[i][j] / _R;
		 _P = ( xU4[i][j] - 0.5 * _R * ( _U*_U + _V*_V ) ) * K_1;
		 _C  = sqrt( K * _P / _R );
		 _jo = _K_1 * _P;	//_P - _C * _C * _R;
		 _jp = _U + ( RU = _P / ( _R * _C ) );
		 _jm = _U -   RU;
		 // right parameters
		 R_ =   U1x[i][j];
		 U_ =   U2x[i][j] / R_;
		 V_ =   U3x[i][j] / R_;
		 P_ = ( U4x[i][j] - 0.5 * R_ * ( U_*U_ + V_*V_ ) ) * K_1;
		 C_  = sqrt( K * P_ / R_ );
		 jo_ = _K_1 * P_; 			//P_ - C_ * C_ * R_;
		 jp_ = U_ + ( RU = P_ / ( R_ * C_ ) );
		 jm_ = U_ -   RU;
		 // linearized procedure
		 U = 0.5 * ( _U + U_ );
		 C = 0.5 * ( _C + C_ );
		 C_p = U + C; C_m = U - C; C_o = U;
		 // J_p and al_p
		 if( C_p > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
		 else           { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		 // J_m and al_m
		 if( C_m > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
		 else           { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		 // J_o and al_o
		 if( C_o > 0. ) { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
		 else           { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		 // parameters on the interface
		 P = ( J_p - J_m ) / ( al_p - al_m );
		 U = J_p - al_p * P;
		 R = ( J_o - P ) / al_o;
		 C = sqrt( K * P / R );
		 // tangential velocities
		 if( U > 0 ) V = _V;
		 else        V = V_;
		 // flux vector
		 Fx1[i][j] = RU = R * U * rx[j];
		 Fx2[i][j] = RU * U + P * rx[j];
		 Fx3[i][j] = RU * V;
		 Fx4[i][j] = RU * ( K_K_1 * P / R + 0.5 * ( U * U + V * V ) );
	  } // end for
   } // end for
   // Y-direction
   for( j = 0; j < HIGG; j++ ) {
	  for( i = 0; i < LEN; i++ ) {
		 // selection
		 if( i < LEN1 ) { if( j < HIG2 ) continue; }
		 else if( i < LEN2 ) { if( j < HIG2 && j > HIG1 ) continue; }
		 // lower
		 _R =   yU1[i][j];
		 _U =   yU2[i][j] / _R;
		 _V =   yU3[i][j] / _R;
		 _P = ( yU4[i][j] - 0.5 * _R * ( _U*_U + _V*_V ) ) * K_1;
		 _C  = sqrt( K * _P / _R );
		 _jo = _K_1 * _P; 			//_P - _C * _C * _R;
		 _jp = _V + ( RU = _P / ( _R * _C ) );
		 _jm = _V -   RU;
		 // upper
		 R_ =   U1y[i][j];
		 U_ =   U2y[i][j] / R_;
		 V_ =   U3y[i][j] / R_;
		 P_ = ( U4y[i][j] - 0.5 * R_ * ( U_*U_ + V_*V_ ) ) * K_1;
		 C_  = sqrt( K * P_ / R_ );
		 jo_ = _K_1 * P_; 			//P_ - C_ * C_ * R_;
		 jp_ = V_ + ( RU = P_ / ( R_ * C_ ) );
		 jm_ = V_ -   RU;
		 // linearized procedure
		 V = 0.5 * ( _V + V_ );
		 C = 0.5 * ( _C + C_ );
		 C_p = V + C; C_m = V - C; C_o = V;
		 // Jp and al_p
		 if( C_p > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
		 else           { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		 // Jm and al_m
		 if( C_m > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
		 else           { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		 // Jo and al_o
		 if( C_o > 0. ) { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
		 else           { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		 // parameters on the interface
		 P = ( J_p - J_m ) / ( al_p - al_m );
		 V = J_p - al_p * P;
		 R = ( J_o - P ) / al_o;
		 C = sqrt( K * P / R );
		 // tangential velocities
		 if( V > 0 ) U = _U;
		 else        U = U_;
		 // flux vector
		 Fy1[i][j] = RU = R * V * ry[j];
		 Fy2[i][j] = RU * U;
		 Fy3[i][j] = RU * V + P * ry[j];
		 Fy4[i][j] = RU * ( K_K_1 * P / R + 0.5 * ( U * U + V * V ) );
	  } // end for
   } // end for

   //-- 1D
   for ( i = 0; i < LEN_; i++ ) {  // except for right end of 1D duct
	  // left state
	  R_1 =   xUU1[i];
	  U_1 =   xUU2[i] / R_1;
	  P_1 = ( xUU3[i] - 0.5 * R_1 * U_1 * U_1 ) * K_1;
	  C_1 = sqrt( K * P_1 / R_1 );
	  _jo = P_1 - C_1 * C_1 * R_1;
	  _jp = U_1 + ( PRC = P_1 / ( R_1 * C_1 ) );
	  _jm = U_1 - PRC;
	  // right state
	  R_2 =   UU1x[i];
	  U_2 =   UU2x[i] / R_2;
	  P_2 = ( UU3x[i] - 0.5 * R_2 * U_2 * U_2 ) * K_1;
	  C_2 = sqrt( K * P_2 / R_2 );
	  jo_ = P_2 - C_2 * C_2 * R_2;
	  jp_ = U_2 + ( PRC = P_2 / ( R_2 * C_2 ) );
	  jm_ = U_2 - PRC;
	  // initial guess
	  U = 0.25  * ( _jp + _jm + jp_ + jm_ );
	  C = _025K * ( _jp - _jm + jp_ - jm_ );
	  // characteristic speeds
	  C_p = U + C, C_m = U - C; C_o = U;
	  // essentially noniterative LCS procedure
	  // proc:
		 // J_p and al_p
	  if( C_p /*U+C*/ > 0. ) { al_p = K_1_2 * ( _jm - _jp ) / _jo; J_p  = _jp; }
	  else                   { al_p = K_1_2 * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		 // J_m and al_m
	  if( C_m /*U-C*/ > 0. ) { al_m = K_1_2 * ( _jp - _jm ) / _jo; J_m  = _jm; }
	  else                   { al_m = K_1_2 * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		// J_o and al_o
	  if( C_o /*U*/ > 0. )   { al_o = _05K * ( _jp - _jm ); al_o = - al_o * al_o; J_o  = _jo; }
	  else                   { al_o = _05K * ( jp_ - jm_ ); al_o = - al_o * al_o; J_o  = jo_; }
		// parameters on the interface
	  P = ( J_p - J_m ) / ( al_p - al_m );
	  U = J_p - al_p * P;
	  R = ( J_o - P ) / al_o;
	  // checking signs of the Ch-velocities - not needed
/*	  C = sqrt( K * P / R );
	  C_p_ = U + C; C_m_ = U - C; C_o_ = U;
	  if( C_p_*C_p < 0 || C_m_*C_m < 0 || C_o_*C_o < 0 ) {
		 C_p = C_p_; C_m = C_m_; C_o = C_o_;
		 goto proc;
	  }*/
	  // fluxes
	  FFx1[i] = R * U;
	  FFx2[i] = R * U * U + P;
	  FFx3[i] = U * ( K_K_1 * P + 0.5 * R * U * U );
   } // end for()

   // right end of 1D duct
   for( j = 0, u1 = u2 = u3 = 0; j < HIG1; j++ ) {
	  u1 += Fx1[LEN1][j];
	  u2 += Fx2[LEN1][j];
	  u3 += Fx4[LEN1][j];
   }
   FFx1[LEN_] = u1 / r_;
   FFx2[LEN_] = u2 / r_;
   FFx3[LEN_] = u3 / r_;

} // end Fluxes()


/**************
*  EVOLUTION  *   new (or intermediate) values of conservative variables
**************/
void Evolution( void )
{
register unsigned i, j, _i, _j;
register real rr;

   switch( Stage ) {
   
	  // 1st stage
	  case 1:
		 // 2D
		 for( j = 1, _j = 0; j < HIGG; j++, _j++ ) {
          rr = r[j];
		    for( i = 1, _i = 0; i < LENN; i++, _i++ ) {
			   if( j <= HIG1 ) { if( i <= LEN1 ) continue; }
			   else if( j <= HIG2 ) { if( i <= LEN2) continue; }
            // pressure
		      U =   U2[i][j] / ( R = U1[i][j] );
		      V =   U3[i][j] / R;
		      P = ( U4[i][j] - 0.5 * R * ( U * U + V * V ) ) * K_1;
            // evolution
            U1p[i][j] = U1[i][j] + ( deltaT_X * ( Fx1[_i][_j] - Fx1[i][_j] )
									        + deltaT_Y * ( Fy1[_i][_j] - Fy1[_i][j] )
                                   ) / rr;
			   U2p[i][j] = U2[i][j] + ( deltaT_X * ( Fx2[_i][_j] - Fx2[i][_j] )
									        + deltaT_Y * ( Fy2[_i][_j] - Fy2[_i][j] )
                                   ) / rr;
			   U3p[i][j] = U3[i][j] + ( deltaT_X * ( Fx3[_i][_j] - Fx3[i][_j] )
									        + deltaT_Y * ( Fy3[_i][_j] - Fy3[_i][j] )
                                   + P * deltaT
                                   ) / rr;
			   U4p[i][j] = U4[i][j] + ( deltaT_X * ( Fx4[_i][_j] - Fx4[i][_j] )
									        + deltaT_Y * ( Fy4[_i][_j] - Fy4[_i][j] )
                                   ) / rr;
			}
		 }
		 // 1D
		 for ( _i = 0, i = 1; i < LENN_; i++, _i++ ) {
			UU1p[i] = UU1[i] + deltaT_X * ( FFx1[_i] - FFx1[i] );
			UU2p[i] = UU2[i] + deltaT_X * ( FFx2[_i] - FFx2[i] );
			UU3p[i] = UU3[i] + deltaT_X * ( FFx3[_i] - FFx3[i] );
		 }
		 /* next stage */
	     Stage = 2;
	     U1_ = U1p,  U2_ = U2p,  U3_ = U3p,  U4_ = U4p;
	     UU1_= UU1p, UU2_= UU2p, UU3_= UU3p;
		 break;

	  // 2nd stage
	  case 2:
		 // 1D
		 for( j = 1, _j = 0; j < HIGG; j++, _j++ ) {
          rr = r[j];
		    for( i = 1, _i = 0; i < LENN; i++, _i++ ) {
			   if( j <= HIG1 ) { if( i <= LEN1 ) continue; }
			   else if( j <= HIG2 ) { if( i <= LEN2) continue; }
            // pressure
		      U =   U2p[i][j] / ( R = U1p[i][j] );
		      V =   U3p[i][j] / R;
		      P = ( U4p[i][j] - 0.5 * R * ( U * U + V * V ) ) * K_1;
            // evolution
            U1p[i][j] = _0_75 * U1[i][j]
                      + _0_25 * ( U1p[i][j] + ( deltaT_X * ( Fx1[_i][_j] - Fx1[i][_j] )
                                              + deltaT_Y * ( Fy1[_i][_j] - Fy1[_i][j] )
                                              ) / rr
                                 );
			   U2p[i][j] = _0_75 * U2[i][j]
                      + _0_25 * ( U2p[i][j] + ( deltaT_X * ( Fx2[_i][_j] - Fx2[i][_j] )
                                              + deltaT_Y * ( Fy2[_i][_j] - Fy2[_i][j] )
                                              ) / rr );
			   U3p[i][j] = _0_75 * U3[i][j]
                      + _0_25 * ( U3p[i][j] + ( deltaT_X * ( Fx3[_i][_j] - Fx3[i][_j] )
                                              + deltaT_Y * ( Fy3[_i][_j] - Fy3[_i][j] )
                                              + P * deltaT
                                              ) / rr );
			   U4p[i][j] = _0_75 * U4[i][j]
                      + _0_25 * ( U4p[i][j] + ( deltaT_X * ( Fx4[_i][_j] - Fx4[i][_j] )
                                              + deltaT_Y * ( Fy4[_i][_j] - Fy4[_i][j] )
                                              ) / rr );
			}
		 }
		 // 1D
		 for ( _i = 0, i = 1; i < LENN_; i++, _i++  ) {
			UU1p[i] = _0_75 * UU1[i] + _0_25 * ( UU1p[i] + deltaT_X * ( FFx1[_i] - FFx1[i] ) );
			UU2p[i] = _0_75 * UU2[i] + _0_25 * ( UU2p[i] + deltaT_X * ( FFx2[_i] - FFx2[i] ) );
			UU3p[i] = _0_75 * UU3[i] + _0_25 * ( UU3p[i] + deltaT_X * ( FFx3[_i] - FFx3[i] ) );
		 }
         /* next stage */
		 Stage = 3;
		 break;

	 // 3rd stage
	 default:
		 // 2D
		 for( j = 1, _j = 0; j < HIGG; j++, _j++ ) {
          rr = r[j];
		    for( i = 1, _i = 0; i < LENN; i++, _i++ ) {
			   if( j <= HIG1 ) { if( i <= LEN1 ) continue; }
			   else if( j <= HIG2 ) { if( i <= LEN2) continue; }
            // pressure
		      U =   U2p[i][j] / ( R = U1p[i][j] );
		      V =   U3p[i][j] / R;
		      P = ( U4p[i][j] - 0.5 * R * ( U * U + V * V ) ) * K_1;
            // evolution
			   U1[i][j] = _0_33 * U1[i][j]
                     + _0_66 * ( U1p[i][j] + ( deltaT_X * ( Fx1[_i][_j] - Fx1[i][_j] )
                                             + deltaT_Y * ( Fy1[_i][_j] - Fy1[_i][j] )
                                             ) / rr );
			   U2[i][j] = _0_33 * U2[i][j]
                     + _0_66 * ( U2p[i][j] + ( deltaT_X * ( Fx2[_i][_j] - Fx2[i][_j] )
                                             + deltaT_Y * ( Fy2[_i][_j] - Fy2[_i][j] )
                                             ) / rr );
			   U3[i][j] = _0_33 * U3[i][j]
                     + _0_66 * ( U3p[i][j] + ( deltaT_X * ( Fx3[_i][_j] - Fx3[i][_j] )
                                             + deltaT_Y * ( Fy3[_i][_j] - Fy3[_i][j] )
                                             + P * deltaT
                                             ) / rr );
			   U4[i][j] = _0_33 * U4[i][j]
                     + _0_66 * ( U4p[i][j] + ( deltaT_X * ( Fx4[_i][_j] - Fx4[i][_j] )
                                             + deltaT_Y * ( Fy4[_i][_j] - Fy4[_i][j] )
                                             ) / rr );
			}
		 }
		 // 1D
		 for ( _i = 0, i = 1; i < LENN_; i++, _i++  ) {
			UU1[i] = _0_33 * UU1[i] + _0_66 * ( UU1p[i] + deltaT_X * ( FFx1[_i] - FFx1[i] ) );
			UU2[i] = _0_33 * UU2[i] + _0_66 * ( UU2p[i] + deltaT_X * ( FFx2[_i] - FFx2[i] ) );
			UU3[i] = _0_33 * UU3[i] + _0_66 * ( UU3p[i] + deltaT_X * ( FFx3[_i] - FFx3[i] ) );
		 }
         /* next stage */		 
	     Stage = 1;
	     U1_ = U1,  U2_ = U2,  U3_ = U3,  U4_ = U4;
	     UU1_= UU1, UU2_= UU2, UU3_= UU3;
		 break;

   } // end switch()

} // end Evolution()


/*********
*  SAVE  *   backup of the solution in binary file
*********/
void Save( void )
{
unsigned i;

	if ( ( pF = fopen( "backup", "wb") ) == NULL ) { puts( "Cannot open file \"backup\""); exit(1);	}
	// 2D
	for ( i = 1; i <= LEN; i++ ) {
	  fwrite( &U1[i][1], HIG * sizeof( real ), 1, pF );
	  fwrite( &U2[i][1], HIG * sizeof( real ), 1, pF );
	  fwrite( &U3[i][1], HIG * sizeof( real ), 1, pF );
	  fwrite( &U4[i][1], HIG * sizeof( real ), 1, pF );
	}
	// 1D
	fwrite( &UU1[1], LEN_ * sizeof( real ), 1, pF );
	fwrite( &UU2[1], LEN_ * sizeof( real ), 1, pF );
	fwrite( &UU3[1], LEN_ * sizeof( real ), 1, pF );
	// ... = step
	fwrite( &step, sizeof( unsigned ), 1, pF );
	fclose( pF );

} // end Save()


/***********
*  OUTPUT  *    output of the flowfield
***********/
void Output( void )
{
unsigned i, j;
char str[80];

   // 2D field
   sprintf( str, "%d", step );
   if ( ( pF = fopen( str, "w" ) ) == NULL ) {
	  printf( "Cannot open file" );
	  exit( 1 );
   }
   fprintf( pF, "%d\n%d\n%f\n%f\n", LEN, HIG, deltaX, deltaY );
   for ( j = 1; j < HIGG; j++ ) {
	  for ( i = 1; i < LENN; i++ ) {
		 if( j <= HIG1 ) {
			if( i <= LEN1 ) {
			   fprintf( pF, "0. \n" );
			   continue;
			}
		 }
		 else if( j <= HIG2 ) {
			if( i <= LEN2) {
			   fprintf( pF, "0. \n" );
			   continue;
			}
		 }
		 fprintf( pF, "%f \n", U1[i][j] );
	  }
   }
   fclose( pF );

   // 1D
   sprintf( str, "%d.dat", step );
   if ( ( pF = fopen( str, "w" ) ) == NULL ) {
	  printf( "Cannot open file" );
	  exit( 1 );
   }
   fprintf( pF, "%d\n%d\n", LEN_ + 2, 3 );
   fprintf( pF, "0. 90000. 0.9 0. \n" );
   for ( i = 1; i < LENN_; i++ ) {
	  R =   UU1[i];
	  U =   UU2[i] / R;
	  P = ( UU3[i] - 0.5 * R * U * U ) * K_1;
	  fprintf( pF,"%1.4f %6.0f %1.4f %3.1f \n", (i-0.5)*deltaX, P, R,
                                               /* what if log10( R ),? */ U );
   }
   fprintf(pF, "%d 210000. 2.1 100. \n", LEN_ );
   fclose( pF );

} // end Output()


/***********
*  Probes  *
***********/
void Probes( void )
{
   // fprintfs data to probes file: P, R, U and V
   U =   U2[i_s][j_s] / ( R = U1[i_s][j_s] );
	V =   U3[i_s][j_s] / R;
   P = ( U4[i_s][j_s] - 0.5 * R * ( U * U + V * V ) ) * K_1;
   fprintf( pFs, "%d %6.3f %1.5f %2.3f %2.3f \n", step, P, R, U, V );
} // end Finalize()


/*************
*  FINALIZE  *
*************/
void Finalize( void )
{
   fclose( pFs ); // closing probes file
} // end Finalize()


/*********
*  MAIN  *
*********/
void main( void )
{
   //--- handler for a SIGINT (Ctrl+C)
   signal( SIGINT, termHandler );
   
   //--- initializaton
   Initialize( );
   
   //---
   t = time(NULL);
   
   //--- time integration loop
   while( 1 ) {
   
     //---
     Probes( );
	 
      //---
	  if( step % outstep == 0 && step ) Output();
	  //--- exit after _numstep_ steps or if [Ctrl+C] pressed...
	  if( step == numstep || !continFlag  ) break; printf( "%lu\n", ++step );
	  
	  //=== 1st stage
	  //--- boundary conditions
	  BoundaryConditions();
	  //--- piecewise-parabolic reconstruction
	  Reconstruction();
	  //--- at the outer sides of cell boundaries
	  BConInterfaces();
	  //--- fluxes at the boundaries
	  Fluxes();
	  //--- intermediate values of conservative variables
	  Evolution();
	  
	  //=== 2nd stage
	  //--- boundary conditions
	  BoundaryConditions();
	  //--- piecewise-parabolic reconstruction
	  Reconstruction();
	  //--- at the outer sides of cell boundaries
	  BConInterfaces();
	  //--- fluxes at the boundaries
	  Fluxes();
	  //--- intermediate values of conservative variables
	  Evolution();
	  
	  //=== 3rd stage
	  //--- boundary conditions
	  BoundaryConditions();
	  //--- piecewise-parabolic reconstruction
	  Reconstruction();
	  //--- at the outer sides of cell boundaries
	  BConInterfaces();
	  //--- fluxes at the boundaries
	  Fluxes();
	  //--- new values of conservative variables
	  Evolution();
	  
   } // end while()
   
   //---
   printf( "Elapsed time is %d sec\n", time(NULL)-t );
   
   //---
   Save();
   
   //--- final output of the flowfield
   Output();
   
   //---
   Finalize();
   
} // end main()

//--- end noise.c